#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Iñigo Serna <inigoserna@gmail.com>
# Time-stamp: <2012-05-14 09:09:51 inigo>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# ebook-thumbnailer.py:
#    create thumbs for .fb2, .epub, .pdf, .mobi and .chm files for GNOME 3
#
# Dependencies: pygtk, cairo, pypoppler, mobi_unpack, python-chm
#
# Instructions: copy binary and thumbnailers definition
#     # cp ebook-thumbnailer.py /usr/local/bin
#     # cp {chm,epub,fb2,mobi}.thumbnailer /usr/share/thumbnailers
# and optionally (not needed as evince already does it):
#     # cp pdf.thumbnailer /usr/share/thumbnailers


import os
import os.path
import sys
from urllib import unquote
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


IMGS_MIMETYPE = ('image/jpg', 'image/jpeg', 'image/png', 'image/bmp')
IMGS_EXT = ('.jpg', '.jpeg', '.png', '.gif')
# THUMBSIZE = (256*2/3, 256)
THUMBSIZE = (128*2/3, 128)


##########################################################################
##### FB2
from base64 import b64decode

def fb2_cover_data(filename):
    NS1 = '{http://www.gribuser.ru/xml/fictionbook/2.0}'
    NS2 = '{http://www.w3.org/1999/xlink}'

    cid = el = None
    tree = ET.parse(filename)
    # coverpage/image
    img = tree.find('{0}description/{0}title-info/{0}coverpage/{0}image'.format(NS1))
    if img is not None:
        if img.attrib.has_key(NS2+'href'):
            cid = img.attrib[NS2+'href'].replace('#', '')
        elif img.attrib.has_key('href'):
            cid = img.attrib['href'].replace('#', '')
    if cid is not None:
        el = tree.find('{0}binary[@id="{1}"]'.format(NS1, cid))
        if el is None or not (el.attrib.has_key('content-type') and el.attrib['content-type'] in IMGS_MIMETYPE):
            el = None
    # else search for valid images and select biggest
    if el is None:
        size = -1
        for elm in tree.findall('{0}binary'.format(NS1)):
            if not (elm.attrib.has_key('content-type') and elm.attrib['content-type'] in IMGS_MIMETYPE):
                continue
            if len(elm.text) > size:
                el, size = elm, len(elm.text)
    return None if el is None else b64decode(el.text)


##########################################################################
##### EPUB
import zipfile

def epub_cover_data(filename):
    NS1 = '{urn:oasis:names:tc:opendocument:xmlns:container}'
    NS2 = '{http://www.idpf.org/2007/opf}'
    NS3 = '{http://www.w3.org/1999/xhtml}'
    NS4 = '{http://www.w3.org/2000/svg}'
    NS5 = '{http://www.w3.org/1999/xlink}'

    coverimg = None
    book = zipfile.ZipFile(filename, 'r')
    root1 = ET.fromstring(book.read('META-INF/container.xml'))
    el = root1.find('{0}rootfiles/{0}rootfile'.format(NS1))
    if el is None or not (el.attrib.has_key('full-path') and el.attrib['media-type']=='application/oebps-package+xml'):
        return None
    content_opf = el.attrib['full-path']
    root2 = ET.fromstring(book.read(content_opf))
    el = root2.find('{0}guide/{0}reference[@type="cover"]'.format(NS2))
    if el is not None and el.attrib.has_key('href'): # cover file found
        cover_file = os.path.normpath(os.path.join(os.path.dirname(content_opf) , el.attrib['href']))
        if cover_file.endswith(IMGS_EXT):
            coverimg = cover_file
        else:
            root3 = ET.fromstring(book.read(unquote(cover_file)))
            imgs = []
            for el in root3.iter():
                if el.tag == NS3+'img' and el.attrib.has_key('src'):
                    imgs.append(el.attrib['src'])
                if el.tag == NS3+'image' or el.tag == NS4+'image':
                    if el.attrib.has_key('href'):
                        imgs.append(el.attrib['href'])
                    elif el.attrib.has_key('xlink:href'):
                        imgs.append(el.attrib['xlink:href'])
                    elif el.attrib.has_key(NS5+'href'):
                        imgs.append(el.attrib[NS5+'href'])
            if len(imgs) > 0:
                coverimg = os.path.normpath(os.path.join(os.path.dirname(cover_file), imgs[0]))
    if coverimg is None:
        el = root2.find('{0}manifest/{0}item[@id="cover"]'.format(NS2))
        if el is not None and (el.attrib.has_key('media-type') and el.attrib['media-type'] in IMGS_MIMETYPE):
            coverimg = os.path.join(os.path.dirname(content_opf), el.attrib['href'])
    if coverimg is None: # no cover found, try biggest image
        size = -1
        for imgfile in [f for f in book.namelist() if f.endswith(IMGS_EXT)]:
            info = book.getinfo(imgfile)
            if info.file_size > size:
                coverimg, size = imgfile, info.file_size
    return None if coverimg is None else book.read(unquote(coverimg), 'r')


##########################################################################
##### PDF: needs pypoppler
import poppler
import cairo

def pdf_cover_data(filename):
    uri = 'file://' + (filename if os.path.isabs(filename) else os.path.abspath(filename))
    doc = poppler.document_new_from_file(uri, '')
    page = doc.get_page(0)
    pw, ph = page.get_size()
    imgsurf = cairo.ImageSurface(cairo.FORMAT_RGB24, THUMBSIZE[0], THUMBSIZE[1])
    ctx = cairo.Context(imgsurf)
    ctx.set_source_rgb(1.0, 1.0, 1.0)
    ctx.paint()
    ctx.scale(THUMBSIZE[0]/pw, THUMBSIZE[1]/ph)
    page.render(ctx)
    stream = gio.MemoryOutputStream() # as faster as StringIO, but module is already in memory
    imgsurf.write_to_png(stream)
    return stream.get_contents()


##########################################################################
##### MOBI/PRC/AZW/AZW4: get mobi_unpack from http://wiki.mobileread.com/wiki/Mobi_unpack
#####                    and change sys.path.append line below
from tempfile import mkdtemp
from shutil import rmtree
import re
sys.path.append('/home/inigo/devel/bidebarrieta/bidebarrieta/3rd_party/MobiUnpack')
from mobi_unpack import unpackBook

class NullDevice():
    def write(self, s):
        pass

def mobi_cover_data(filename):
    NS1 = '{http://www.idpf.org/2007/opf}'

    # uncompress book, get and parse .opf files for cover
    destdir = mkdtemp()
    sys.stdout = NullDevice()
    unpackBook(filename, destdir)
    sys.stdout = sys.__stdout__
    img_files = []
    for root, dirs, files in os.walk(destdir):
        for f in files:
            if f.endswith(('.opf', '.OPF')):
                img = None
                opf_file = os.path.join(root, f)
                try:
                    tree = ET.parse(opf_file)
                    el = tree.find('{0}metadata/{0}meta[@name="EmbeddedCover"]'.format(NS1))
                    if el is None:
                        el = tree.find('{0}metadata/{0}meta[@name="Cover ThumbNail Image"]'.format(NS1))
                    if el is not None and el.attrib.has_key('content'):
                        img = el.attrib['content']
                except:
                    match = re.search(r'<meta name="EmbeddedCover" content="(.*)"\s*/>', open(opf_file).read(), re.I|re.M)
                    if match is not None:
                        img = match.group(1)
                if img is not None:
                    img_files.append(os.path.join(os.path.dirname(opf_file), img))
    # if imgs => return first, else: walk path => return biggest
    if len(img_files) > 0:
        imgcover = img_files[0]
    else:
        for root, dirs, files in os.walk(destdir):
            for f in files:
                if f.endswith(IMGS_EXT):
                    ff = os.path.join(root, f)
                    img_files.append((os.path.getsize(ff), ff))
        imgcover = sorted(img_files)[-1][1] if len(img_files)>0 else None
    imgdata = None if imgcover is None else open(unquote(imgcover)).read()
    rmtree(destdir)
    return imgdata


##########################################################################
##### CHM: needs python-chm module
#####      code adapted from http://ubuntuforums.org/showthread.php?t=1159569
from chm import chm

def chm_cover_data(filename):
    def chm_getfile(chmfile, fname):
        (ret, ui) = chmfile.ResolveObject(fname)
        if ret == 1:
            return (0, '')
        return chmfile.RetrieveObject(ui)

    imgdata = None
    chmfile = chm.CHMFile()
    if chmfile.LoadCHM(filename) == 0:
        return None
    size, data = chm_getfile(chmfile, chmfile.home)
    if size > 0 and chmfile.home.endswith(IMGS_EXT):
        imgdata = data
    else:
        base = chmfile.home.rpartition('/')[0] + '/'
        imgs_size = []
        data = ' '.join(data.split()) # only one line
        imgs = re.findall('<img.*?src="([^"]+)"[^>]*>', data, re.I|re.M|re.S)
        imgs.extend(re.findall("<img.*?src='([^']+)'[^>]*>", data, re.I|re.M|re.S))
        for img in imgs:
            fname = unquote(base + img)
            size, data = chm_getfile(chmfile, fname)
            if imgs_size.count((size, fname)) == 0:
                imgs_size.append((size, fname))
        if len(imgs_size) > 0:
            size, data = chm_getfile(chmfile, sorted(imgs_size)[-1][1])
            if size > 0:
                imgdata = data
    chmfile.CloseCHM()
    return imgdata


##########################################################################
import urllib
import gio
import gtk.gdk
# import StringIO
# import Image


def create_thumb(file_ebook, file_thumb):
    try:
        if file_ebook.endswith('.fb2'):
            img_data = fb2_cover_data(file_ebook)
        elif file_ebook.endswith('.epub'):
            img_data = epub_cover_data(file_ebook)
        elif file_ebook.endswith('.pdf'):
            img_data = pdf_cover_data(file_ebook)
        elif file_ebook.endswith(('.mobi', '.prc', '.azw', '.azw4')):
            img_data = mobi_cover_data(file_ebook)
        elif file_ebook.endswith('.chm'):
            img_data = chm_cover_data(file_ebook)
        else:
            print 'WARNING: file "%s" not supported' % file_ebook
            return
    except:
        print 'ERROR: %s' % file_ebook
        return
    if img_data is not None:
        # print '=> %s: cover' % file_ebook
        # im = Image.open(StringIO(img_data))
        # im.thumbnail(THUMBSIZE, Image.ANTIALIAS) # Image.BILINEAR
        # im.save(filename+'.png', 'PNG')
        # im.save(filename+'.jpeg', 'JPEG')
        stream = gio.memory_input_stream_new_from_data(img_data)
        pb = gtk.gdk.pixbuf_new_from_stream_at_scale(stream, THUMBSIZE[0], THUMBSIZE[1])
        pb.save(file_thumb, 'png')
    # else:
    #     print '=> %s: no cover' % file_ebook


def main():
   if len(sys.argv) == 3:
        file_ebook = urllib.url2pathname(sys.argv[1]).replace('file://', '')
        file_thumb = urllib.url2pathname(sys.argv[2]).replace('file://', '')
        create_thumb(file_ebook, file_thumb)


if __name__ == '__main__':
    main()


##########################################################################
